package at.ac.uibk.gitsearch.service;

import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.security.core.userdetails.User;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import at.ac.uibk.gitsearch.domain.UserWatchList;
import at.ac.uibk.gitsearch.repository.UserWatchListRepository;
import at.ac.uibk.gitsearch.security.AuthoritiesConstants.AuthoritiesConstantEnum;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider;
import at.ac.uibk.gitsearch.service.dto.UserWatchListDTO;
import at.ac.uibk.gitsearch.service.mapper.UserWatchListMapper;
import at.ac.uibk.gitsearch.web.rest.errors.BadRequestAlertException;

/**
 * Service Implementation for managing {@link UserWatchList}.
 */
@Service
@Transactional
public class UserWatchListService {

    private final Logger log = LoggerFactory.getLogger(UserWatchListService.class);

    private final UserWatchListRepository userWatchListRepository;

    private final UserWatchListMapper userWatchListMapper;

    private final TokenProvider tokenProvider;

    public UserWatchListService(UserWatchListRepository userWatchListRepository, UserWatchListMapper userWatchListMapper, TokenProvider tokenProvider) {
        this.userWatchListRepository = userWatchListRepository;
        this.userWatchListMapper = userWatchListMapper;
        this.tokenProvider = tokenProvider;
    }

    /**
     * checks the access to this watchlist for the current user.
     * @param watchlistId the id of the watchlist
     * @param description some description of the request
     * @throws BadRequestAlertException if access not allowed
     */
	public void checkAccessToWatchList(final Long watchlistId, final String description) throws IllegalAccessError {
		final Optional<User> currentPrincipal = tokenProvider.getCurrentPrincipal();
		if(currentPrincipal.isEmpty()) {
    		log.warn("Cannot find a principal for watchlist {} for exercise {}", watchlistId, description);
            throw new IllegalAccessError("Cannot find a principal");
    	}
		if(currentPrincipal.get().getAuthorities().contains(AuthoritiesConstantEnum.ADMIN.getGrantedAuthority())) {
			return; // ADMIN is always allowed
		}
    	final Optional<UserWatchListDTO> watchListO = findOne(watchlistId);
    	if(watchListO.isEmpty()) {
    		log.warn("Cannot find watchlist for : {} for exercise {}", watchlistId, description);
            throw new IllegalAccessError("Cannot find watchlist");
    	}
    	final boolean isAccessible = watchListO.get().getUserIdLogin().equals(currentPrincipal.get().getUsername());
    	if(!isAccessible) {
    		log.warn("watchlist {} does not belong to current user",  watchListO.get().getName());
            throw new IllegalAccessError("watchlist does not belong to current User");
    	}
	}

    /**
     * Save a userWatchList.
     *
     * @param userWatchListDTO the entity to save.
     * @return the persisted entity.
     */
    public UserWatchListDTO save(UserWatchListDTO userWatchListDTO) {
        log.debug("Request to save UserWatchList : {}", userWatchListDTO);
        UserWatchList userWatchList = userWatchListMapper.toEntity(userWatchListDTO);
        userWatchList = userWatchListRepository.save(userWatchList);
        UserWatchListDTO result = userWatchListMapper.toDto(userWatchList);
        return result;
    }

    /**
     * Get all the userWatchLists.
     *
     * @param pageable the pagination information.
     * @return the list of entities.
     */
    @Transactional(readOnly = true)
    public Page<UserWatchListDTO> findAll(Pageable pageable) {
        log.debug("Request to get all UserWatchLists");
        return userWatchListRepository.findAll(pageable)
            .map(userWatchListMapper::toDto);
    }

    /**
     * Get all the userWatchLists.
     *
     * @param pageable the pagination information.
     * @return the list of entities.
     */
    @Transactional(readOnly = true)
    public List<UserWatchListDTO> findByUserIdIsCurrentUser() {
        log.debug("Request to get all UserWatchLists");
        return userWatchListRepository.findByUserIdIsCurrentUser().stream()
            .map(userWatchListMapper::toDto).collect(Collectors.toList());
    }


    /**
     * Get one userWatchList by id.
     *
     * @param id the id of the entity.
     * @return the entity.
     */
    @Transactional(readOnly = true)
    public Optional<UserWatchListDTO> findOne(Long id) {
        log.debug("Request to get UserWatchList : {}", id);
        return userWatchListRepository.findById(id)
            .map(userWatchListMapper::toDto);
    }

    /**
     * Delete the userWatchList by id.
     *
     * @param id the id of the entity.
     */
    public void delete(Long id) {
        log.debug("Request to delete UserWatchList : {}", id);
        userWatchListRepository.deleteById(id);
    }

    /**
     * Search for the userWatchList corresponding to the query.
     *
     * @param query the query of the search.
     * @param pageable the pagination information.
     * @return the list of entities.
     */
    @Transactional(readOnly = true)
    public Page<UserWatchListDTO> search(String query, Pageable pageable) {
        log.debug("Request to search for a page of UserWatchLists for query {}", query);
        final List<Sort.Order> cleanSortOrder = pageable.getSort().get().filter(order -> !order.getProperty().equals( "_score")).collect(Collectors.toList());
        Pageable cleanedPageRequest = PageRequest.of(pageable.getPageNumber(), pageable.getPageSize(), Sort.by(cleanSortOrder));
        return userWatchListRepository.search(query, cleanedPageRequest)
            .map(userWatchListMapper::toDto);
    }

    /**
     * Search for the userWatchList corresponding to the query.
     *
     * @param query the query of the search.
     * @param pageable the pagination information.
     * @return the list of entities.
     */
    @Transactional(readOnly = true)
    public Page<UserWatchListDTO> searchForCurrentUser(String query, Pageable pageable) {
        log.debug("Request to search for a page of CurrentUserWatchLists for query {}", query);
        final List<Sort.Order> cleanSortOrder = pageable.getSort().get().filter(order -> !order.getProperty().equals( "_score")).collect(Collectors.toList());
        Pageable cleanedPageRequest = PageRequest.of(pageable.getPageNumber(), pageable.getPageSize(), Sort.by(cleanSortOrder));
        return userWatchListRepository.searchForCurrentUser(query, cleanedPageRequest)
            .map(userWatchListMapper::toDto);
    }

}
