package at.ac.uibk.gitsearch.web.rest;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.List;
import java.util.Optional;

import javax.validation.Valid;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpHeaders;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.security.core.userdetails.User;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.PutMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.servlet.support.ServletUriComponentsBuilder;

import at.ac.uibk.gitsearch.security.SecurityUtils;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider;
import at.ac.uibk.gitsearch.service.SearchService;
import at.ac.uibk.gitsearch.service.UserService;
import at.ac.uibk.gitsearch.service.UserWatchListQueryService;
import at.ac.uibk.gitsearch.service.UserWatchListService;
import at.ac.uibk.gitsearch.service.WatchListEntryService;
import at.ac.uibk.gitsearch.service.dto.SearchInputDTO;
import at.ac.uibk.gitsearch.service.dto.SearchResultsDTO;
import at.ac.uibk.gitsearch.service.dto.UserWatchListCriteria;
import at.ac.uibk.gitsearch.service.dto.UserWatchListDTO;
import at.ac.uibk.gitsearch.service.dto.WatchListEntryDTO;
import at.ac.uibk.gitsearch.web.rest.errors.BadRequestAlertException;
import io.github.jhipster.service.filter.LongFilter;
import io.github.jhipster.web.util.HeaderUtil;
import io.github.jhipster.web.util.PaginationUtil;
import io.github.jhipster.web.util.ResponseUtil;

/**
 * REST controller for managing {@link at.ac.uibk.gitsearch.domain.UserWatchList}.
 */
@RestController
@RequestMapping("/api")
public class UserWatchListResource {

    private final Logger log = LoggerFactory.getLogger(UserWatchListResource.class);

    private static final String ENTITY_NAME = "userWatchList";

//  @Value("${jhipster.clientApp.name}") // warning: inconsistent rename of gitSearchApp to gitSearchV2App :-(
	private final static String applicationName = "gitsearchApp";

    private final UserService userService;
    private final SearchService searchService;
    
    private final UserWatchListService userWatchListService;
    private final WatchListEntryService watchListEntryService;

    private final UserWatchListQueryService userWatchListQueryService;

    private final TokenProvider tokenProvider;
    
    public UserWatchListResource(UserWatchListService userWatchListService,
    		UserWatchListQueryService userWatchListQueryService, 
    		TokenProvider tokenProvider, UserService userService, 
    		SearchService searchService,
    		WatchListEntryService watchListEntryService) {
        this.userWatchListService = userWatchListService;
        this.userWatchListQueryService = userWatchListQueryService;
        this.tokenProvider = tokenProvider;
        this.userService = userService;
        this.watchListEntryService = watchListEntryService;
        this.searchService = searchService;
    }

    /**
     * {@code POST  /user-watch-lists} : Create a new userWatchList.
     *
     * @param userWatchListDTO the userWatchListDTO to create.
     * @return the {@link ResponseEntity} with status {@code 201 (Created)} and with body the new userWatchListDTO, or with status {@code 400 (Bad Request)} if the userWatchList has already an ID.
     * @throws URISyntaxException if the Location URI syntax is incorrect.
     */
    @PostMapping("/user-watch-lists")
    public ResponseEntity<UserWatchListDTO> createUserWatchList(@Valid @RequestBody UserWatchListDTO userWatchListDTO) throws URISyntaxException {
        log.debug("REST request to save UserWatchList : {}", userWatchListDTO);
       if(tokenProvider.getCurrentPrincipal().isEmpty()) {
           log.debug("unknown user cannot create watchlist {}", userWatchListDTO.getName());
           return ResponseEntity.badRequest().build();
       }
       if(userWatchListDTO.getUserIdId() == null) {
    	   Optional<at.ac.uibk.gitsearch.domain.User> u =  userService.getUserWithAuthoritiesByLogin(SecurityUtils.getCurrentUserLogin().get());
    	   if(u.isEmpty()) {
               log.debug(" user not found, cannot create watchlist {}", userWatchListDTO.getName());
               return ResponseEntity.badRequest().build();
    	   }
    	   userWatchListDTO.setUserIdId(u.get().getId());
    	   userWatchListDTO.setUserIdLogin(u.get().getLastName());
       }
        if (userWatchListDTO.getId() != null) {
            throw new BadRequestAlertException("A new userWatchList cannot already have an ID", ENTITY_NAME, "idexists");
        }
        UserWatchListDTO result = userWatchListService.save(userWatchListDTO);
        return ResponseEntity.created(new URI("/api/user-watch-lists/" + result.getId()))
            .headers(HeaderUtil.createEntityCreationAlert(applicationName, true, ENTITY_NAME, result.getId().toString()))
            .body(result);
    }

    /**
     * {@code POST  /user-watch-lists} : Create a new userWatchList.
     *
     * @param userWatchListDTO the userWatchListDTO to create.
     * @return the {@link ResponseEntity} with status {@code 201 (Created)} and with body the new userWatchListDTO, or with status {@code 400 (Bad Request)} if the userWatchList has already an ID.
     * @throws URISyntaxException if the Location URI syntax is incorrect.
     */
    @PostMapping("/currentuser-watch-lists")
    public ResponseEntity<UserWatchListDTO> createWatchListForCurrentUser(@Valid @RequestBody UserWatchListDTO userWatchListDTO) throws URISyntaxException {
        log.debug("REST request to save UserWatchList for current User: {}", userWatchListDTO);
        if (userWatchListDTO.getId() != null) {
            throw new BadRequestAlertException("A new userWatchList cannot already have an ID", ENTITY_NAME, "idexists");
        }
        final Optional<User> currentUser = tokenProvider.getCurrentPrincipal();
        if(currentUser.isEmpty()) {
        	log.warn("Cannot create a watchlist {} for unknown user", userWatchListDTO.getName());
            throw new BadRequestAlertException("Cannot create a watchlist for unknown user", ENTITY_NAME, "unknownUser");
        }
        User springUser = currentUser.get();
        final Optional<at.ac.uibk.gitsearch.domain.User> userWithAuthoritiesByLogin = userService.getUserWithAuthoritiesByLogin(springUser.getUsername());
        if(userWithAuthoritiesByLogin.isEmpty()) {
        	log.warn("Cannot create a watchlist {} for unknown user", springUser.getUsername());
            throw new BadRequestAlertException("Cannot create a watchlist for unknown user", ENTITY_NAME, "unknownUser");
        }
        userWatchListDTO.setUserIdLogin(springUser.getUsername());
		userWatchListDTO.setUserIdId(userWithAuthoritiesByLogin.get().getId());
        UserWatchListDTO result = userWatchListService.save(userWatchListDTO);
        return ResponseEntity.created(new URI("/api/user-watch-lists/" + result.getId()))
            .headers(HeaderUtil.createEntityCreationAlert(applicationName, true, ENTITY_NAME, result.getId().toString()))
            .body(result);
    }

    /**
     * {@code PUT  /user-watch-lists} : Updates an existing userWatchList.
     *
     * @param userWatchListDTO the userWatchListDTO to update.
     * @return the {@link ResponseEntity} with status {@code 200 (OK)} and with body the updated userWatchListDTO,
     * or with status {@code 400 (Bad Request)} if the userWatchListDTO is not valid,
     * or with status {@code 500 (Internal Server Error)} if the userWatchListDTO couldn't be updated.
     * @throws URISyntaxException if the Location URI syntax is incorrect.
     */
    @PutMapping("/user-watch-lists")
    @PreAuthorize("hasAnyRole('ADMIN')")
    public ResponseEntity<UserWatchListDTO> updateUserWatchList(@Valid @RequestBody UserWatchListDTO userWatchListDTO) throws URISyntaxException {
        log.debug("REST request to update UserWatchList : {}", userWatchListDTO);
        if (userWatchListDTO.getId() == null) {
            throw new BadRequestAlertException("Invalid id", ENTITY_NAME, "idnull");
        }
        UserWatchListDTO result = userWatchListService.save(userWatchListDTO);
        return ResponseEntity.ok()
            .headers(HeaderUtil.createEntityUpdateAlert(applicationName, true, ENTITY_NAME, userWatchListDTO.getId().toString()))
            .body(result);
    }

    /**
     * {@code GET  /user-watch-lists} : get all the userWatchLists.
     *
     * @param pageable the pagination information.
     * @param criteria the criteria which the requested entities should match.
     * @return the {@link ResponseEntity} with status {@code 200 (OK)} and the list of userWatchLists in body.
     */
    @GetMapping("/user-watch-lists")
    @PreAuthorize("hasAnyRole('ADMIN')")
    public ResponseEntity<List<UserWatchListDTO>> getAllUserWatchLists(UserWatchListCriteria criteria, Pageable pageable) {
        log.debug("REST request to get UserWatchLists by criteria: {}", criteria);
        Page<UserWatchListDTO> page = userWatchListQueryService.findByCriteria(criteria, pageable);
        HttpHeaders headers = PaginationUtil.generatePaginationHttpHeaders(ServletUriComponentsBuilder.fromCurrentRequest(), page);
        return ResponseEntity.ok().headers(headers).body(page.getContent());
    }

    /**
     * {@code GET  /user-watch-lists} : get all the userWatchLists.
     *
     * @param pageable the pagination information.
     * @param criteria the criteria which the requested entities should match.
     * @return the {@link ResponseEntity} with status {@code 200 (OK)} and the list of userWatchLists in body.
     */
    @GetMapping("/currentuser-watch-lists")
    public ResponseEntity<List<UserWatchListDTO>> getUserWatchListsOfCurrentUser(UserWatchListCriteria criteria, Pageable pageable) {
        log.debug("REST request to get CurrentUserWatchLists by criteria: {}", criteria);
        final Optional<at.ac.uibk.gitsearch.domain.User> currentUser = userService.getUserWithAuthorities();
        if(currentUser.isEmpty()) {
            return ResponseEntity.notFound().headers(HeaderUtil.createFailureAlert(applicationName, true, "User", "not found",
                    "There was an error: User not found.")).build();        
        }
        final LongFilter userIdFilter = new LongFilter();
        userIdFilter.setEquals(currentUser.get().getId());
        
		criteria.setUserIdId(userIdFilter);
        Page<UserWatchListDTO> page = userWatchListQueryService.findByCriteria(criteria, pageable);
        HttpHeaders headers = PaginationUtil.generatePaginationHttpHeaders(ServletUriComponentsBuilder.fromCurrentRequest(), page);
        return ResponseEntity.ok().headers(headers).body(page.getContent());
    }

    /**
     * {@code GET  /user-watch-lists/count} : count all the userWatchLists.
     *
     * @param criteria the criteria which the requested entities should match.
     * @return the {@link ResponseEntity} with status {@code 200 (OK)} and the count in body.
     */
    @GetMapping("/user-watch-lists/count")
    @PreAuthorize("hasAnyRole('ADMIN')")
    public ResponseEntity<Long> countUserWatchLists(UserWatchListCriteria criteria) {
        log.debug("REST request to count UserWatchLists by criteria: {}", criteria);
        return ResponseEntity.ok().body(userWatchListQueryService.countByCriteria(criteria));
    }

    /**
     * {@code GET  /user-watch-lists/:id} : get the "id" userWatchList.
     *
     * @param id the id of the userWatchListDTO to retrieve.
     * @return the {@link ResponseEntity} with status {@code 200 (OK)} and with body the userWatchListDTO, or with status {@code 404 (Not Found)}.
     */
    @GetMapping("/user-watch-lists/{id}")
    @PreAuthorize("hasAnyRole('ADMIN')")
    public ResponseEntity<UserWatchListDTO> getUserWatchList(@PathVariable Long id) {
        log.debug("REST request to get UserWatchList : {}", id);
        Optional<UserWatchListDTO> userWatchListDTO = userWatchListService.findOne(id);
        return ResponseUtil.wrapOrNotFound(userWatchListDTO);
    }

    /**
     * {@code DELETE  /user-watch-lists/:id} : delete the "id" userWatchList.
     *
     * @param id the id of the userWatchListDTO to delete.
     * @return the {@link ResponseEntity} with status {@code 204 (NO_CONTENT)}.
     */
    @DeleteMapping("/user-watch-lists/{id}")
    public ResponseEntity<Void> deleteUserWatchList(@PathVariable Long id) {
    	userWatchListService.checkAccessToWatchList(id, "searchExercisesOnWatchlist");
        log.debug("REST request to delete UserWatchList : {}", id);
        userWatchListService.delete(id);
        return ResponseEntity.noContent().headers(HeaderUtil.createEntityDeletionAlert(applicationName, true, ENTITY_NAME, id.toString())).build();
    }

    /**
     * {@code SEARCH  /_search/user-watch-lists?query=:query} : search for the userWatchList corresponding
     * to the query.
     *
     * @param query the query of the userWatchList search.
     * @param pageable the pagination information.
     * @return the result of the search.
     */
    @GetMapping("/_search/user-watch-lists")
    @PreAuthorize("hasAnyRole('ADMIN')")
    public ResponseEntity<List<UserWatchListDTO>> searchUserWatchLists(@RequestParam String query, Pageable pageable) {
        log.debug("REST request to search for a page of UserWatchLists for query {}", query);
        Page<UserWatchListDTO> page = userWatchListService.search(query, pageable);
        HttpHeaders headers = PaginationUtil.generatePaginationHttpHeaders(ServletUriComponentsBuilder.fromCurrentRequest(), page);
        return ResponseEntity.ok().headers(headers).body(page.getContent());
        }
    
    /**
     * {@code SEARCH  /_search/user-watch-lists?query=:query} : search for the userWatchList of the current user corresponding
     * to the query.
     *
     * @param query the query of the userWatchList search.
     * @param pageable the pagination information.
     * @return the result of the search.
     */
    @GetMapping("/_search/currentuser-watch-lists")
    public ResponseEntity<List<UserWatchListDTO>> searchCurrentUserWatchLists(@RequestParam String query, Pageable pageable) {
        log.debug("REST request to search for a current user and page of UserWatchLists for query {}", query);
        Page<UserWatchListDTO> page = userWatchListService.searchForCurrentUser(query, pageable);
        HttpHeaders headers = PaginationUtil.generatePaginationHttpHeaders(ServletUriComponentsBuilder.fromCurrentRequest(), page);
        return ResponseEntity.ok().headers(headers).body(page.getContent());
        }

    @PostMapping("/currentuser-watch-lists/{id}")
    public SearchResultsDTO
    searchExercisesOnWatchlist(@PathVariable Long id, @RequestBody Integer page) throws IOException {
    	if(page == null) page = 0;
    	userWatchListService.checkAccessToWatchList(id, "searchExercisesOnWatchlist");
    	final List<WatchListEntryDTO> entriesForWatchlist = watchListEntryService.getEntriesForWatchlist(id);
        return searchService.watchListResultPage(entriesForWatchlist.stream().map(entry -> entry.getExerciseId()), page, SearchInputDTO.PAGE_SIZE);
    }

}
